import sys
from typing import Any, Union

if sys.version_info < (3, 8):
    from typing_extensions import Protocol, runtime_checkable
else:
    from typing import Protocol, runtime_checkable


__all__ = ["ICommitable", "IStrSerializer"]


@runtime_checkable
class ICommitable(Protocol):
    def commit(self) -> None:
        ...


@runtime_checkable
class IStrSerializer(Protocol):
    def dumps(self, obj: Any, *args, **kwargs) -> Union[str, bytes]:
        ...

    def loads(self, s: Union[str, bytes], *args, **kwargs) -> Any:
        ...



import typing as _t

import marshmallow.fields as _fields
from marshmallow import Schema
from marshmallow.fields import (
    UUID,
    Bool,
    Boolean,
    Date,
    DateTime,
    Decimal,
    Dict,
    Email,
    Field,
    Float,
    Int,
    Integer,
    IPv4,
    IPv6,
    List,
    Number,
    Str,
    String,
    Time,
    TimeDelta,
    Url,
)


class Nested(_fields.Nested):
    def __init__(self, dict: _t.Dict[str, _t.Any], *args, **kwargs):
        _fields.Nested.__init__(self, Schema.from_dict(dict), *args, **kwargs)



import sys
from typing import Callable, Iterable

from nekoite_be_core.view import FrameworkAdapterBase, ViewBase, Response
from nekoite_be_core.util import unify_to_list
from nekoite_be_core.errors import BEError

from marshmallow import ValidationError

try:
    from flask import Flask, request, Response as FlaskResponse
except ModuleNotFoundError:
    print("Module flask is needed to use flask adapter.", file=sys.stderr)
    exit(1)


__all__ = ["default_routefunc_generator", "FlaskAdapter"]


def default_routefunc_generator(
    app: Flask, route: str, view: ViewBase, methods: Iterable[str]
):
    # TODO: support async.
    @app.route(route, endpoint=route, methods=methods)
    def _f_():
        if request.method != "GET" and view._is_json_body_ and not request.is_json:
            raise RuntimeError("Not json post request")
        if request.method in ("PUT", "POST"):
            req = request.json
        else:
            req = request.args
        try:
            req = view.parse_request(req)
        except ValidationError as e:
            # TODO: add customized error processing
            return {"code": 400, "message": str(e)}
        code = 0
        message = ""
        ret = None
        try:
            # if inspect.iscoroutinefunction(view.handle_req):
            #     ret = await view.handle_req(req)
            # else:
            ret = view.handle_req(req)
        except BEError as e:
            code = e.code
            message = e.message if e.message else ""
        except Exception as e:
            code = 500
            # message = traceback.format_exc(1)
            message = str(e)
            raise e
        if isinstance(ret, Response):
            code = ret.code
            message = ret.message
            data = ret.data
        elif isinstance(ret, dict):
            try:
                data = view.parse_response(ret)
            except ValidationError as e:
                code = 500
        elif isinstance(ret, FlaskResponse):
            return ret
        else:
            data = ret

        return {"code": code, "message": message, "data": data}


class FlaskAdapter(FrameworkAdapterBase):
    def __init__(
        self,
        app: Flask,
        routefunc_generator: Callable[
            [Flask, str, ViewBase, Iterable[str]], None
        ] = default_routefunc_generator,
    ) -> None:
        super().__init__()
        self._app = app
        self._routefunc_generator = routefunc_generator

    def register_route(self, route: str, view: ViewBase):
        methods = unify_to_list(view.methods)
        self._routefunc_generator(self._app, route, view, methods)



from typing import Any, Dict, Union, Iterable, Type, Optional
from abc import ABC, abstractmethod
from marshmallow import Schema


__all__ = ["Response", "ViewBase", "FrameworkAdapterBase", "register_routes"]


class Response:
    def __init__(self, code: int, message: Optional[str], data: Any) -> None:
        self._code = code
        self._message = message
        self._data = data

    @property
    def code(self) -> int:
        return self._code

    @property
    def message(self) -> Optional[str]:
        return self._message

    @property
    def data(self) -> Any:
        return self._data


class ViewBase(ABC):
    _is_json_body_ = True

    request_schema: Dict[str, Any]
    response_schema: Dict[str, Any]
    methods: Union[Iterable[str], str]

    def __init__(self):
        self._req_sch: Schema = (
            Schema.from_dict(self.request_schema)()
            if getattr(self, "request_schema", None) is not None
            else None
        )
        self._resp_sch: Schema = (
            Schema.from_dict(self.response_schema)()
            if getattr(self, "response_schema", None) is not None
            else None
        )

    def parse_request(self, req: Dict[str, Any]) -> Dict[str, Any]:
        if not self._req_sch:
            return req
        return self._req_sch.load(req)

    @abstractmethod
    def handle_req(self, req: Dict[str, Any]) -> Any:
        raise NotImplementedError()

    def parse_response(self, resp: Any) -> Any:
        if not isinstance(resp, dict) or not self._resp_sch:
            return resp
        return self._resp_sch.dump(resp)


class FrameworkAdapterBase(ABC):
    @abstractmethod
    def register_route(self, route: str, view: ViewBase):
        raise NotImplementedError()


def register_routes(
    adapter: FrameworkAdapterBase,
    routes: Dict[str, Type[ViewBase]],
    strict_trailing_slash: bool = True,
):
    for route, view in routes.items():
        adapter.register_route(route, view())
        if not strict_trailing_slash:
            another_rt = route
            if route.endswith("/"):
                another_rt = route[:-1]
            else:
                another_rt = route + "/"
            if another_rt not in routes:
                adapter.register_route(another_rt, view())
